/***************************************************************************/
/* sexp.h								   */
/*									   */
/* General S-expression structure definitions for SPKI code.		   */
/***************************************************************************/

#if !defined(__SEXP_H__)
#define __SEXP_H__

#include "abbrev.h"

/***************************************************************************/
/* byte string structure -- with optional display hint			   */
/***************************************************************************/

typedef struct _bytes_ {	/* byte-string */
  struct _bytes_ *nxt ;		/* next byte string, for free.c */
  short refcnt ;		/* reference count (drop when 0) */
  short drop_cont ;		/* non-0 means internal strings were carved */
  int lu ;			/* lookup value */
  UL d_lth ;			/* length of the display hint */
  CP disp ;			/* display hint */
  UL v_lth ;			/* length of the byte string itself */
  UCP val ;			/* byte string */
} Bytes, *pBytes ;

/***************************************************************************/
/* Types of S-expression elements					   */
/***************************************************************************/

#define S_KW (1)		/* 1st term: recognized keyword */
#define S_UKW (2)		/* 1st term: unknown keyword */
#define S_BS (3)		/* other term: general byte string */
#define S_SEXP (4)		/* other term: sub-expression */

/***************************************************************************/
/* S-expression element							   */
/***************************************************************************/

typedef struct _sexp_ {		/* general S-expression */
  struct _sexp_ *nxt ;		/* next S-expression, for free.c */
  LI refcnt ;			/* reference count (drop when 0) */
  struct _sexp_ *term ;		/* next term in the expression */
  /* the expression bytes and length are NULL if not 1st term */
  UCP expb ;			/* expression bytes, if in contiguous memory */
  LI explth ;			/* expression length, if in contig. mem. */
  int typ ;			/* type of this term */
  union {
    int k ;			/* S_KW: recognized keyword */
    pBytes b ;			/* S_UKW or S_BS: byte string  */
    struct _sexp_ *s ;		/* S_SEXP: sub-expression */
  } u ;
} Sexp, *pSexp ;

#endif /* __SEXP_H__ */
